<?php

if (!defined('ABSPATH')) exit;

class gdpos_core_search {
    private $_power = false;

    private $_query = array();
    private $_default = array(
        'types' => 'posts',
        'status' => 'all',
        'forums' => 'all',
        'forums_selection' => array(),
        'forums_subforums' => 'yes',
        'users_input' => '',
        'users_select' => '',
        'orderby' => 'date',
        'order' => 'DESC',
        'period' => 'any',
        'tags' => ''
    );

    private $_valid = array(
        'types' => array('types'), 
        'status' => array('status'), 
        'forums' => array('forums', 'forums_selection', 'forums_subforums'), 
        'users' => array('users_input', 'users_select'), 
        'period' => array('period'), 
        'tags' => array('tags')
    );

    public function __construct() {
        add_action('gdpos_plugin_init', array($this, 'init'), 1);
        add_action('gdpos_parse_query_string', array($this, 'parse'));

        add_filter('gdpos_search_query_parse_for_forums_selection', array($this, 'parse_absint_array'));
        add_filter('gdpos_search_query_parse_for_tags', array($this, 'parse_tags'));
    }

    public function init() {
        if (gdpos_settings()->get('power_search_query_arg') && isset($_GET['power-search'])) {
            $this->_power = true;
        }
    }

    public function parse() {
        $this->_power = true;

        $values = $this->available_query_args();

        foreach ($values as $key) {
            $real = 'gdpos_search_'.$key;

            if (isset($_GET[$real]) && !empty($_GET[$real])) {
                $raw = $_GET[$real];
                $processed = '';

                if (!is_array($raw)) {
                    $raw = urldecode($raw);

                    $processed = d4p_sanitize_basic($raw);
                } else {
                    $raw = array_map('urldecode', $raw);

                    $processed = d4p_sanitize_basic_array($raw);
                }

                $processed = apply_filters('gdpos_search_query_parse_for_'.$key, $processed, $raw);

                if ($this->_default[$key] !== $processed) {
                    $this->_query[$key] = $processed;
                }
            }
        }

        if (isset($this->_query['forums'])) {
            if ($this->_query['forums'] == 'forums') {
                if (isset($this->_query['forums_selection']) && empty($this->_query['forums_selection'])) {
                    unset($this->_query['forums']);
                }
            } else if ($this->_query['forums'] == 'subscribe') {
                if (!is_user_logged_in()) {
                    unset($this->_query['forums']);
                }
            }
        }

        $this->_query = apply_filters('gdpos_search_loaded_query', $this->_query);
    }

    public function is_power() {
        return $this->_power;
    }

    public function get_query() {
        return $this->_query;
    }

    public function get_query_value($name, $default = '') {
        return isset($this->_query[$name]) ? $this->_query[$name] : (
               isset($this->_default[$name]) ? $this->_default[$name] : $default);
    }

    public function available_query_args() {
        $available = array('orderby', 'order');

        foreach ($this->_valid as $key => $keys) {
            if (gdpos()->form()->show($key)) {
                $available = array_merge($available, $keys);
            }
        }

        return $available;
    }

    public function parse_absint_array($processed) {
        return array_map('absint', $processed);
    }

    public function parse_tags($processed) {
        $tags = explode(',', $processed);
        $tags = array_map('trim', $tags);
        $tags = array_filter($tags);

        return join(',', $tags);
    }

    public function search_query_args($args) {
        global $wpdb;

        $args['_gdpos_search_where'] = 'everywhere';
        add_filter('posts_search', array($this, 'override_search_query'), 100, 2);

        $orderby = $this->get_query_value('orderby', 'date');
        if ($orderby != 'date') {
            $args['orderby'] = $orderby;
        }

        $order = $this->get_query_value('order', 'DESC');
        if ($order != 'date') {
            $args['order'] = $order;
        }

        $types = $this->get_query_value('types');
        switch ($types) {
            case 'topic':
                $args['post_type'] = bbp_get_topic_post_type();
                break;
            case 'reply':
                $args['post_type'] = bbp_get_reply_post_type();
                break;
            default:
            case 'posts':
                $args['post_type'] = array(bbp_get_topic_post_type(), bbp_get_reply_post_type());
                break;
        }

        $status = $this->get_query_value('status');
        if ($status != 'all') {
            if ($args['post_type'] === bbp_get_topic_post_type()) {
                $args['post_status'] = $status;
            } else {
                $args['_gdpos_search_status'] = $status;

                add_filter('posts_join_request', array($this, 'override_join_request_status'), 160, 2);
            }
        }

        $forums = $this->get_query_value('forums');
        $ids = array();

        $_exclude = array();
        $_include = gdpos()->bbpress()->allowed_forums_ids();

        if (empty($_include)) {
            $_exclude = gdpos()->bbpress()->disallowed_forums_ids();
        }

        switch ($forums) {
            case 'forums':
                $list = $this->get_query_value('forums_selection', array());
                $subs = $this->get_query_value('forums_subforums') == 'yes';

                if ($subs) {
                    $ids = gdpos()->cache()->forums_with_subforums($list);
                } else {
                    $ids = $list;
                }
                break;
            case 'public':
                $ids = gdpos()->cache()->public_forums();
                break;
            case 'subscribed':
                $ids = bbp_get_user_subscribed_forum_ids(bbp_get_current_user_id());
                break;
        }

        if (!empty($ids)) {
            if (!empty($_include)) {
                $ids = array_intersect($ids, $_include);
            } else if (!empty($_exclude)) {
                $ids = array_diff($ids, $_exclude);
            }

            if (empty($ids)) {
                $ids = array(-1);
            }

            if (!isset($args['meta_query'])) {
                $args['meta_query'] = array();
            }

            $args['meta_query'][] = array(
                'compare' => 'IN',
                'type' => 'UNSIGNED',
                'key' => '_bbp_forum_id',
                'value' => $ids
            );
        } else {
            if (!empty($_include)) {
                if (!isset($args['meta_query'])) {
                    $args['meta_query'] = array();
                }

                $args['meta_query'][] = array(
                    'compare' => 'IN',
                    'type' => 'UNSIGNED',
                    'key' => '_bbp_forum_id',
                    'value' => $_include
                );
            } else if (!empty($_exclude)) {
                if (!isset($args['meta_query'])) {
                    $args['meta_query'] = array();
                }

                $args['meta_query'][] = array(
                    'compare' => 'NOT IN',
                    'type' => 'UNSIGNED',
                    'key' => '_bbp_forum_id',
                    'value' => $_exclude
                );
            }
        }

        $users_input = $this->get_query_value('users_input');
        if (!empty($users_input)) {
            $args['_gdpos_search_users_input'] = $users_input;

            add_filter('posts_join_request', array($this, 'override_join_request'), 100, 2);
            add_filter('posts_where_request', array($this, 'override_where_request'), 100, 2);
        }

        $users_select = $this->get_query_value('users_select');
        if (!empty($users_select)) {
            $user_id = absint($users_select);

            $args['author'] = $user_id;
        }

        $tags = $this->get_query_value('tags');
        if (!empty($tags)) {
            $terms = array();
            $tags = explode(',', $tags);

            foreach ($tags as $tag) {
                $raw = get_terms(array('taxonomy' => bbp_get_topic_tag_tax_id(), 'search' => $tag, 'fields' => 'tt_ids'));

                $terms[] = empty($raw) ? '0' : join(',', $raw);
            }

            $args['_gdpos_search_tags'] = $terms;

            add_filter('posts_join_request', array($this, 'override_join_request_tags'), 110, 2);
        }

        $period = $this->get_query_value('period');
        if ($period != 'any') {
            $dates = array();

            switch ($period) {
                case 'day':
                    $dates[] = "{$wpdb->posts}.post_date >= DATE_SUB(CURDATE(), INTERVAL 1 DAY)";
                    break;
                case 'week':
                    $dates[] = "{$wpdb->posts}.post_date >= DATE_SUB(CURDATE(), INTERVAL 7 DAY)";
                    break;
                case 'month':
                    $dates[] = "{$wpdb->posts}.post_date >= DATE_SUB(CURDATE(), INTERVAL 1 MONTH)";
                    break;
                case 'half':
                    $dates[] = "{$wpdb->posts}.post_date >= DATE_SUB(CURDATE(), INTERVAL 6 MONTH)";
                    break;
                case 'year':
                    $dates[] = "{$wpdb->posts}.post_date >= DATE_SUB(CURDATE(), INTERVAL 1 YEAR)";
                    break;
            }

            $args['_gdpos_search_period'] = $dates;

            add_filter('posts_where_request', array($this, 'override_where_request_period'), 120, 2);
        }

        return apply_filters('gdpos_search_query_args', $args);
    }

    public function override_join_request($join, $wp_query) {
        global $wpdb;

        $q = $wp_query->query_vars;

        if (isset($q['_gdpos_search_users_input'])) {
            remove_filter('posts_join_request', array($this, 'override_join_request'), 100, 2);

            $join.= " LEFT JOIN $wpdb->users gdpos_users ON ($wpdb->posts.post_author = gdpos_users.ID) ";
        }

        return $join;
    }

    public function override_where_request($where, $wp_query) {
        global $wpdb;

        $q = $wp_query->query_vars;

        if (isset($q['_gdpos_search_users_input'])) {
            remove_filter('posts_where_request', array($this, 'override_where_request'), 100, 2);

            $like = '%'.$wpdb->esc_like($q['_gdpos_search_users_input']).'%';

            $where.= $wpdb->prepare(" AND (gdpos_users.display_name LIKE %s OR gdpos_users.user_login LIKE %s) ", $like, $like);
        }

        return $where;
    }

    public function override_join_request_tags($join, $wp_query) {
        global $wpdb;

        $q = $wp_query->query_vars;

        if (isset($q['_gdpos_search_tags'])) {
            remove_filter('posts_join_request', array($this, 'override_join_request_tags'), 110, 2);

            $terms = $q['_gdpos_search_tags'];

            $join.= " INNER JOIN {$wpdb->postmeta} gdpos_mt ON ({$wpdb->posts}.ID = gdpos_mt.post_id AND gdpos_mt.meta_key = '_bbp_topic_id') ";

            $i = 1;
            foreach ($terms as $t) {
                $join.= " INNER JOIN $wpdb->term_relationships gdpos_tr_$i ON (gdpos_tr_$i.object_id = CAST(gdpos_mt.meta_value as UNSIGNED) AND gdpos_tr_$i.term_taxonomy_id IN ($t)) ";

                $i++;
            }
        }

        return $join;
    }

    public function override_join_request_status($join, $wp_query) {
        global $wpdb;

        $q = $wp_query->query_vars;

        if (isset($q['_gdpos_search_status'])) {
            remove_filter('posts_join_request', array($this, 'override_join_request_status'), 150, 2);

            $status = $q['_gdpos_search_status'];

            $join = " INNER JOIN {$wpdb->postmeta} gdpos_status ON gdpos_status.post_id = {$wpdb->posts}.ID AND gdpos_status.meta_key = '_bbp_topic_id' INNER JOIN {$wpdb->posts} gdpos_status_topic ON gdpos_status_topic.ID = CAST(gdpos_status.meta_value AS UNSIGNED) AND gdpos_status_topic.post_status = '$status'";
        }

        return $join;
    }

    public function override_where_request_period($where, $wp_query) {
        $q = $wp_query->query_vars;

        if (isset($q['_gdpos_search_period'])) {
            remove_filter('posts_where_request', array($this, 'override_where_request'), 100, 2);

            $periods = $q['_gdpos_search_period'];

            $where.= " AND (".join(" AND ", $periods).") ";
        }

        return $where;
    }

    public function override_search_query($search, $wp_query) {
        global $wpdb;

        $q = $wp_query->query_vars;

        if (isset($q['_gdpos_search_where'])) {
            $where = $q['_gdpos_search_where'];

            if (count($q['search_terms']) == 1 &&
                in_array($q['search_terms'][0], array('-', '_'))) {
                return '';
            }

            $search = '';
            $searchand = '';
            $n = !empty($q['exact']) ? '' : '%';

            $exclusion_prefix = apply_filters('wp_query_search_exclusion_prefix', '-');

            foreach ($q['search_terms'] as $term) {
                $exclude = $exclusion_prefix && ($exclusion_prefix === substr($term, 0, 1));

                if ($exclude) {
                    $like_op = 'NOT LIKE';
                    $andor_op = 'AND';
                    $term = substr($term, 1);
                } else {
                    $like_op = 'LIKE';
                    $andor_op = 'OR';
                }

                $like = $n.$wpdb->esc_like($term).$n;

                $search.= $wpdb->prepare("{$searchand}(({$wpdb->posts}.post_title $like_op %s) $andor_op ({$wpdb->posts}.post_excerpt $like_op %s) $andor_op ({$wpdb->posts}.post_content $like_op %s))", $like, $like, $like);

                $searchand = ' AND ';
            }

            if (!empty($search)) {
                $search = " AND ({$search}) ";

                if (!is_user_logged_in()) {
                    $search.= " AND ({$wpdb->posts}.post_password = '') ";
                }
            }
        }

        return $search;
    }
}
